# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import List

import click
from vdk.api.plugin.hook_markers import hookspec
from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.internal.core.context import CoreContext


class InternalHookSpecs:
    """
    These are the hook specs that are used in internal implementation or testing of vdk-core.
    And we are not considering them useful and impactful enough to be made public.
    """

    @hookspec(firstresult=True)
    def vdk_main(
        self,
        plugin_registry: IPluginRegistry,
        root_command: click.Group,
        command_line_args: List,
    ) -> int:
        """
        Called for running the main mltethod of the CLI.
        Default implementation does register hook specs and implementations, initialize the CLI and executes it.

        :param plugin_registry: the Plugin Registry which can be used to add hook specs or implementations
        :param root_command: the root CLI command
        :param command_line_args: the command line args
        :return: the exit code. Immediately after that sys.exit(exit_code) is called.
        """
        pass

    @hookspec(firstresult=True)
    def vdk_cli_execute(
        self,
        root_command: click.Command,
        command_line_args: List,
        program_name: str,
        core_context: CoreContext,
    ) -> int:
        """

        :param root_command: the (root) click command that will be executed and run the application
        :param command_line_args: the arguments with which it is started
        :param program_name: the name of the program.
        :param core_context: the initialized core context of the CLI
        :return: the exit code with which the CLI execution finishes
        In case of error it may also throw an exception
        """



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import pathlib
from dataclasses import dataclass
from typing import cast

from vdk.api.job_input import IJobArguments
from vdk.api.job_input import IJobInput
from vdk.api.job_input import ITemplate
from vdk.api.plugin.connection_hook_spec import (
    ConnectionHookSpec,
)
from vdk.api.plugin.plugin_input import IIngesterRegistry
from vdk.api.plugin.plugin_input import IManagedConnectionRegistry
from vdk.api.plugin.plugin_input import IPropertiesRegistry
from vdk.api.plugin.plugin_input import ITemplateRegistry
from vdk.internal.builtin_plugins.connection.impl.router import ManagedConnectionRouter
from vdk.internal.builtin_plugins.ingestion.ingester_router import IngesterRouter
from vdk.internal.builtin_plugins.job_properties.properties_router import (
    PropertiesRouter,
)
from vdk.internal.builtin_plugins.run.step import StepBuilder
from vdk.internal.core.context import CoreContext


# TODO: split into mutable and immutable job context (e.g JobContextBuilder and JobContext)
# JobContext is mutable only until initializing phase. After init phase only state store should be mutable.


@dataclass()
class JobContext:
    """
    Keep context about a Data Job run (execution).
    It is used by plugins (during job_initialize phase) to setup the data job with the necessary context.
    Things like what database connections are possible, what are the steps of the job and their order,
    how will they be executed, what is the templates that are registered to be used and so on.
    """

    # the name of the job
    name: str
    # Job directory
    job_directory: pathlib.Path
    # the context in which the job is initialized
    # Give access to configuration of the job and to state store.
    # Plugins can keep their own context in the statestore and access it later
    # Common job context - like setps, connections, templates, arguments is kept in JobContext directly
    # instead of in Statestore.
    core_context: CoreContext
    # the steps to be executed composed as a tree
    step_builder: StepBuilder
    # the supported connections
    connections: IManagedConnectionRegistry
    # initialize templates
    templates: ITemplateRegistry
    # job arguments passed by user for current execution
    job_args: IJobArguments
    # initialized job input
    job_input: IJobInput
    # initialize ingestion
    ingester: IIngesterRegistry
    # initialize ingestion
    properties: IPropertiesRegistry

    def __init__(
        self,
        name: str,
        job_directory: pathlib.Path,
        core_context: CoreContext,
        job_args: IJobArguments,
        templates: ITemplateRegistry,
    ):
        self.name = name
        self.job_directory = job_directory
        self.core_context = core_context
        self.step_builder = StepBuilder()
        self.job_args = job_args

        self.connections = ManagedConnectionRouter(
            core_context.configuration,
            cast(ConnectionHookSpec, core_context.plugin_registry.hook()),
        )
        self.templates = cast(ITemplateRegistry, templates)
        self.ingester = IngesterRouter(core_context.configuration, core_context.state)

        self.properties = PropertiesRouter(
            job_name=self.name, cfg=core_context.configuration
        )

        from vdk.internal.builtin_plugins.run.job_input import JobInput

        self.job_input = JobInput(
            cast(ManagedConnectionRouter, self.connections),
            self.core_context.state,
            cast(PropertiesRouter, self.properties),
            self.job_args,
            cast(ITemplate, self.templates),
            cast(IngesterRouter, self.ingester),
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.ingestion import ingester_configuration
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core.config import ConfigurationBuilder


class IngesterConfigurationPlugin:
    """
    Configuration plugin to ensure all plugin configurations, relevant to ingestion
    plugins are ready when the plugins are loaded.
    """

    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        """
        Here we define the configuration settings needed for all ingestion
        plugins with reasonable defaults.
        """
        # Plugin-related configurations
        config_builder.add(
            key="INGEST_METHOD_DEFAULT",
            default_value=None,
            description="Default Ingestion method to be used.",
        )
        config_builder.add(
            key="INGEST_TARGET_DEFAULT",
            default_value=None,
            description="Default Ingestion target to be used.",
        )

        # Configure ingestion specific environment variables
        ingester_configuration.add_definitions(config_builder=config_builder)

    @hookimpl
    def finalize_job(self, context: JobContext) -> None:
        context.ingester.close()



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import configparser
import pathlib
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping

from vdk.internal.core.config import convert_value_to_type_of_default_type


class JobConfigKeys(str, Enum):
    TEAM = "team"
    NOTIFIED_ON_JOB_FAILURE_USER_ERROR = "notified_on_job_failure_user_error"
    NOTIFIED_ON_JOB_FAILURE_PLATFORM_ERROR = "notified_on_job_failure_platform_error"
    NOTIFIED_ON_JOB_SUCCESS = "notified_on_job_success"
    NOTIFIED_ON_JOB_DEPLOY = "notified_on_job_deploy"
    ENABLE_ATTEMPT_NOTIFICATIONS = "enable_attempt_notifications"
    DB_DEFAULT_TYPE = "db_default_type"


class JobConfig:
    """
    User facing configuration of a Data Job.
    For more see the user wiki
    """

    def __init__(self, data_job_path: pathlib.Path):
        self._config_ini = configparser.ConfigParser()
        self._config_ini.read(str(data_job_path.joinpath("config.ini")))

    def get_team(self) -> str:
        """
        Specified which is the team that owns the data job.

        Teams is a way to group data jobs that belong together and are managed by same people.
        It will control who has permissions to manages the data job.
        """
        return self._get_value("owner", JobConfigKeys.TEAM.value)

    def get_contacts_notified_on_job_failure_user_error(self) -> List[str]:
        """
        List of email addresses to be notified on job execution failure caused by user code or user configuration problem
        """
        return self._get_contacts(
            JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_USER_ERROR.value
        )

    def get_contacts_notified_on_job_failure_platform_error(self) -> List[str]:
        """
        List of email addresses to be notified on job execution failure caused by a platform problem.
        """
        return self._get_contacts(
            JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_PLATFORM_ERROR.value
        )

    def get_contacts_notified_on_job_success(self) -> List[str]:
        """
        List of email addresses to be notified on job execution success.
        """
        return self._get_contacts(JobConfigKeys.NOTIFIED_ON_JOB_SUCCESS.value)

    def get_contacts_notified_on_job_deploy(self) -> List[str]:
        """
        List of email addresses to be notified of job deployment outcome.
        """
        return self._get_contacts(JobConfigKeys.NOTIFIED_ON_JOB_DEPLOY.value)

    def get_enable_attempt_notifications(self) -> bool:
        """
        Flag to enable or disable the email notifications on a single attempt
        (in other words each automatic retry will sent an email).
        """
        enable_attempt_notif_value = self._get_value(
            "contacts", JobConfigKeys.ENABLE_ATTEMPT_NOTIFICATIONS.value
        )
        if (
            enable_attempt_notif_value == ""
        ):  # if the parameter hasn't been set in config.ini, we default to False
            return False
        else:
            return convert_value_to_type_of_default_type(
                JobConfigKeys.ENABLE_ATTEMPT_NOTIFICATIONS.value,
                enable_attempt_notif_value,
                False,
            )

    def get_db_type(self) -> str:
        """
        The default database type used by the job (IMPALA, PRESTO)
        """
        return self._get_value("job", JobConfigKeys.DB_DEFAULT_TYPE.value)

    def get_vdk_options(self) -> Dict[str, str]:
        if self._config_ini.has_section("vdk"):
            return dict(self._config_ini["vdk"])
        else:
            return {}

    def _get_value(self, section, key):
        if self._config_ini.has_option(section, key):
            return self._config_ini.get(section, key)
        return ""

    def _get_contacts(self, key):
        contacts_str = self._get_value("contacts", key).strip()
        if contacts_str:
            return [x.strip() for x in contacts_str.split(";")]
        return []



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import cast
from typing import Generic
from typing import Optional
from typing import TypeVar

T = TypeVar("T")


@dataclass(frozen=True)
class StoreKey(Generic[T]):
    """StoreKey is an object used as a key to a Store.

    A StoreKey is associated with the type T of the value of the key.
    """

    key: str


class ImmutableStoreKey(StoreKey[T]):
    """
    StoreKey associated with value which is considered immutable once set.

    Trying to set it again will produce error and raise exception.
    The value itself may or may not be immutable.
    """

    pass


# TODO(minor): should be enum but in python 3.7 and 3.8 cannot declare enum with generic data types
# TODO(minor): also if enum perhaps we do not need string value e.g OP_ID = StoreKey[str]() would be better.
class CommonStoreKeys:
    """
    StoreKeys for a common (standard) attributes used during lifetime of the CLI.
    """

    """
    OpID is similar to trace ID in open tracing.
    It enable tracing multiple operations across difference services and datasets
    """
    OP_ID: StoreKey[str] = StoreKey[str]("vdk.op_id")
    """
    Execution ID is the current id of the current execution.
    It can be passed externally (e.g in case of external re-tries are  considered part of 1 execution)
    """
    EXECUTION_ID: StoreKey[str] = StoreKey[str]("vdk.execution_id")
    """
    It may be the same as Execution_ID unless re-tries per executions are supported
    """
    ATTEMPT_ID: StoreKey[str] = StoreKey[str]("vdk.attempt_id")

    """
    When the CLI (and hence job) execution has been started
    """
    START_TIME: StoreKey[datetime] = StoreKey[datetime]("vdk.start_time")
    """
    When the CLI (and hence job) execution has finished
    """
    END_TIME: StoreKey[datetime] = StoreKey[datetime]("vdk.end_time")

    """
    Build information about the CLI:
    """
    VDK_VERSION: StoreKey[str] = ImmutableStoreKey[str]("vdk.vdk_version")


class StateStore:
    """
    StateStore is a type-safe heterogeneous mapping that allows keys and value types to be defined to keep state of app.

    It is meant to keeps state of the current CLI execution.
    Contains data like opId, executionID, start time, end time, steps executed, etc.
    Telemetry, logging, monitoring, etc. facilities read data from this class and re-model the data to suit their
    needs.

    If a module wants to store data in this StateStore, it creates StoreKeys:

    .. highlight:: python
    .. code-block:: python

        key_with_string_value = StoreKey[str]()
        key_with_bool_value = StoreKey[bool]()

    To store data:

    .. code-block:: python

        # Value type must match the key.
        store[key_with_string_value] = "value"
        store[key_with_bool_value] = True

    To retrieve the data:

    .. code-block:: python

        # The type of string_value is str.
        string_value = store[key_with_string_value]
        # The type of bool_value is bool.
        bool_value = store[key_with_bool_value]

    """

    def __init__(self, store: dict[StoreKey[Any], object] = None) -> None:
        self.__store = {} if store is None else store

    def __setitem__(self, key: StoreKey[T], value: T) -> None:
        """Set a value for key."""
        self.__check_for_immutable_key_existance(key)
        self.__store[key] = value

    def __check_for_immutable_key_existance(self, key: StoreKey[T]) -> None:
        if isinstance(key, ImmutableStoreKey):
            if key in self.__store:
                raise AttributeError(
                    f"Cannot update value for key {key} because key is immutable "
                    f"and value with that key already exists."
                )

    def __getitem__(self, key: StoreKey[T]) -> T:
        """Get the value for key.

        Raises ``KeyError`` if the key wasn't set before.
        """
        return cast(T, self.__store[key])

    def get(self, key: StoreKey[T], default: T = None) -> T | None:
        """Get the value for key, or return default if the key wasn't set before."""
        try:
            return self[key]
        except KeyError:
            return default

    def set(self, key: StoreKey[T], value: T) -> None:
        """Set a value for the given key."""
        self[key] = value

    def __delitem__(self, key: StoreKey[T]) -> None:
        """Delete the value for key.

        Raises ``KeyError`` if the key wasn't set before.
        Raises ``AttributeError`` if the key is set and is immutable.
        """
        self.__check_for_immutable_key_existance(key)
        del self.__store[key]

    def __contains__(self, key: StoreKey[T]) -> bool:
        """Return whether key was set."""
        return key in self.__store

    def __repr__(self) -> str:
        """Return string representation of the state store"""
        return self.__store.__repr__()

    def clone(self) -> StateStore:
        """Returns a new statestore object that is a copy of this instance.
        Clone performs shallow copy - only the top-level objects are duplicated and the lower levels contain references
        """
        store_copy = self.__store.copy()
        return StateStore(store_copy)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pathlib
from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from vdk.internal.builtin_plugins.connection.managed_connection_base import (
    ManagedConnectionBase,
)

# represents any compatible PEP249 Connection class (most of python database clients)
PEP249Connection = Any


class IPropertiesServiceClient:
    """
    Client interface for reading and writing job properties.
    Different backend services can implement it so properties can be stored in different places
    (e.g databae, inmemory, vault)
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def read_properties(self, job_name: str, team_name: str) -> Dict:
        """
        Read properties from the backend - returns a dictionary of all properties for the given data job.
        :param team_name: The name of the team the job belongs to
        :param job_name: the name of the job
        """
        pass

    @abstractmethod
    def write_properties(self, job_name: str, team_name: str, properties: Dict) -> None:
        """
        Write properties to the backend service. It overwrites all properties completely.
        :param team_name: The name of the team the job belongs to
        :param job_name: the name of the job
        :param properties: dictionary with the new job's properties
        """
        pass


IPropertiesFactory = Callable[[], IPropertiesServiceClient]


class IPropertiesRegistry(ABC):
    """
    Properties registry to enable register implementation Properties service
    """

    @abstractmethod
    def set_properties_factory_method(
        self,
        properties_type: str,
        properties_factory: IPropertiesFactory,
    ) -> None:
        """
        Register properties implementation backend.
        Properties API enable keeping state and secrets of a data job.
        Default implementation is in-memory so it's strongly recommended to install vdk-properties plugin
        which provides API based properties implementation

        IPropertiesServiceClient is used as provides logic of how properties are persisted
        and is part of implementation of IProperties interface.
        See

        :param properties_type: str
                string identifying the type of properties being set.
                This enable users to configure what type of properties implemenation to use.
        :param properties_factory: Callable[[], IPropertiesServiceClient]
               method or interface that returns correctly initialized implementation of IPropertiesServiceClient interface
        """
        pass


class ITemplateRegistry:
    """
    Template registry interface for adding new data jobs as templates.

    see job_input.ITemplate for how the template will be used by Data Engineers.

    Templates functionally enables to compose data jobs within themselves.
    For example common platform team can define common data jobs as templates and use them within other data jobs.
    This interface is used by plugin developers to register new templates.

    """

    # TODO: add supported arguments so that they can be validated
    @abstractmethod
    def add_template(self, name: str, template_directory: pathlib.Path):
        """
        Register a new template.
        You usually would want to register template with job_initialize plugin hook.

        For example:

        .. code-block:: python

            @hookimpl
            def initialize_job(self, context: JobContext) -> None:
                context.templates.add_template(
                    "template_name", pathlib.Path(template_name.job_module.__file__)
                )

        :param name: The name of the template . This name will be used by users of vdk in their data jobs when invoking it
        :param template_directory: the directory which contains the implementation of the template. This is a data job which accept arguments.
        VDK will automatically run the data job internally and return result (success or not to the user.). See ITemplate for more details.
        :return:
        """
        pass


class IManagedConnectionRegistry:
    def add_open_connection_factory_method(
        self,
        dbtype: str,
        open_connection_func: Callable[
            [], Union[ManagedConnectionBase, PEP249Connection]
        ],
    ) -> None:
        """

        Register new connection factory method.
        By registering the connection it will be made available by data engineers using job_input.

        For example:

        .. code-block:: python

            @hookimpl(tryfirst=True)
            def vdk_configure(config_builder: ConfigurationBuilder) -> None:
                # Here we define what configuration settings are needed to initialize the connection
                config_builder.add(
                    key="host",  description="The host we need to connect ."
                )
                config_builder.add(
                    key="port", description="The port to connect to"
                )

            @hookimpl
            def initialize_job(context: JobContext) -> None:
                context.connections.add_open_connection_factory_method(
                    "BIG_QUERY", lambda: dbapi.connect(host=cfg.get_value('host'), post=cfg.get_value('port')
                )

        :param dbtype: the name of the database connection - e.g: redshift, impala, presto, big-query, postgres, etc.
        :param open_connection_func:
            The function must implement the code necessary to open a new connection that includes getting all configuration necessary like url, user, etc.
            The function can return either original PEP 249 (DBAPI) Connection from the python library for the given database
            or can wrap the connection into ManagedConnection by subclassing . The second method is only necessary
            if plugin does some pre / post processing and want to decorate some methods (e.g to collecto stats for each query)

        :return:
        """
        pass


class IIngesterPlugin:
    def ingest_payload(
        self,
        payload: List[dict],
        destination_table: Optional[str],
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
    ):
        """
        Do the actual ingestion of the payload

        :param payload: List[dict]
            The payloads to be ingested. Depending on the number of payloads to be
            processed, there might 0 or many dict objects. Each dict object is a
            separate payload.
            Note: The memory size of the list is dependent on the
            payload_size_bytes_threshold attribute.
        :param destination_table: Optional[string]
            (Optional) The name of the table, where the data should be ingested into.
            This argument is optional, and needs to be considered only when the name
            of the destination table is not included in the payload itself.
        :param target: Optional[string]
            (Optional) Used to identify where the data should be ingested into.
                Specifies a data source and its destination database.
                The values for this parameter can be in the format
                `<some-data-source_and-db-table>`, or as a URL.
                Example: http://example.com/<some-api>/<data-source_and-db-table>
            This parameter does not need to be used, in case the
            `INGEST_TARGET_DEFAULT` environment variable is set. This can be made by
            plugins, which may set default value, or it can be overwritten by users.
        :param collection_id: string
            (Optional) An identifier to indicate that data from different method
            invocations belong to same collection. Defaults to "data_job_name|OpID",
            meaning all method invocations from a data job run will belong to the
            same collection.
        """
        pass


class IIngesterRegistry:
    """
    Ingester Registry interface for adding different types of ingester plugins.
    """

    def add_ingester_factory_method(
        self,
        method: str,
        ingester_plugin: Callable[[], IIngesterPlugin],
    ) -> None:
        """
        Register new ingester objects.
        By registering ingester objects, data engineers would be able to use the
        same vdk setup with multiple ingester plugins, and direct where the data
        should be ingested by only specifying the `method` when calling
        send_object_for_ingestion()/send_tabular_data_for_ingestion().

        :param method: str
               Ingestion method to be associated with a plugin.
               For example:
                    method='file' -> for ingest to file plugin
                    method='http' -> for ingest over http plugin
                    method='kafka' -> for ingest to kafka endpoint plugin
               The value for the method parameter will be converted to lowercase and
               stored in this way with the aim of preventing issues where a user has
               set a variable to the correct string for a particular method, but the
               case is wrong.
        :param ingester_plugin: Callable[[], IIngesterPlugin]
               The ingester must implement the ingestion mechanism indicated by the
               method parameter.
        """
        pass




RELEASE_VERSION='0.0.405166560'
BUILD_DATE='Tue Nov  9 12:18:50 UTC 2021'
BUILD_MACHINE_INFO='Linux runner-t4krtqay-project-28359933-concurrent-09pqkx 5.4.129-62.227.amzn2.x86_64 #1 SMP Wed Jul 7 00:08:43 UTC 2021 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='1763689788'
GIT_COMMIT_SHA='22e4b736294872e03326b14bbca45add91002d4c'
GIT_BRANCH='main'



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import Container
from typing import Optional

from vdk.api.plugin.hook_markers import hookspec
from vdk.internal.builtin_plugins.connection.decoration_cursor import DecorationCursor
from vdk.internal.builtin_plugins.connection.recovery_cursor import RecoveryCursor


class ConnectionHookSpec:
    """
    These are hook specifications that enable plugins to hook
        at PEP249Connection and PEP249Cursor events during execution.
    Sequence of evaluation:
        1. db_connection_validate_operation
        2. db_connection_decorate_operation
        * cursor is executed
            ** in case of recovery needed -> db_connection_recover_operation

    Use Cases
        db_connection_validate_operation: if valid return  else raise exception
        db_connection_decorate_operation: ..set_operation( prefix + operation)
        db_connection_recover_operation: ..get_retries()..retry_operation()..
    """

    @hookspec
    def db_connection_validate_operation(
        self, operation: str, parameters: Optional[Container]
    ) -> None:
        """
        Validates the query and parameters.

        For example:
        @hookimpl
        db_connection_validate_operation(operation, parameters):
            if(_max_query_length_limit_exceeded(operation, parameters)):
                raise Exception("max query length limit exceeded")

        :param operation: str
            Database operation - a SQL query or command to be executed.
        :param parameters: Optional[Container]
            Parameters may be provided as sequence or mapping and will be bound to variables in the operation.
            Variables are specified in a database-specific notation. See chosen database documentation for details.
        :return:
        """
        pass

    @hookspec
    def db_connection_decorate_operation(
        self, decoration_cursor: DecorationCursor
    ) -> None:
        """
        Curates the operation and parameters.
        If an exception is raised, the cursor execution will be suspended.

        For example:
        @hookimpl
        db_connection_decorate_operation(decoration_cursor: DecorationCursor):
            managed_operation = decoration_cursor.get_managed_operation()
            managed_operation.set_operation("prefix" +
                                             managed_operation.get_operation())

        :param decoration_cursor: DecorationCursor
            A PEP249Cursor implementation purposed for query and parameters decoration.
            Provides operation details and tooling.
        :return:
        """
        pass

    @hookspec
    def db_connection_recover_operation(self, recovery_cursor: RecoveryCursor) -> None:
        """
        Recovers the operation initiated. Retries made number is auto-incremented.
        If no exception is raised, cursor is considered successfully executed.

        For example:
        @hookimpl
        db_connection_recover_operation(recovery_cursor: RecoveryCursor):
            while recovery_cursor.get_retries() < MAX_RETRIES:
                try:
                    recovery_cursor.execute("helper query")
                    recovery_cursor.retry_operation()
                    return
                except:
                    time.sleep(3 * recovery_cursor.get_retries()) # backoff
            raise recovery_cursor.get_exception()

        :param recovery_cursor: RecoveryCursor
            A PEP249Cursor implementation purposed for query and parameters recovery.
            Provides operation details and tooling.
        :return:
        """
        pass
